use crate::background_loader::MetricsLogReadingRequester;
use crate::config::{GraphRenderingHints, GraphRequest};
use bare_metrics_core::structures::UnixTimestampMilliseconds;
use eframe::egui::{
    Color32, Frame as EguiFrame, PointerButton, Pos2, Rect, Sense, Stroke, TextStyle, Ui, Vec2,
};
use std::ops::{Mul, RangeInclusive};

/// Renderer for a bare metrics graph.
/// Should support rendering both line graphs and histograms.
/// Input interactions:
///   - drag-select an area to zoom in
///   - hover along line to show tooltip with statistics
///   - double click to zoom back out?
///   - right click for context menu with more options (this might be handled at a layer higher
///     up; we'll see...)
pub struct Graph {}

pub struct GraphState {
    pub x_scale: f64,
    pub y_min: f64,
    pub y_max: f64,
    pub x_last_time: UnixTimestampMilliseconds,
}

#[derive(Copy, Clone, Debug)]
pub struct GraphTransform {
    pub scale_x: f64,
    pub scale_y: f64,
    pub offs_x: f64,
    pub offs_y: f64,
}

/// Inspired by egui's RectTransform but uses f64 because f32 don't have enough precision
/// to work well with unix timestamps...
impl GraphTransform {
    pub fn from_ranges_and_display_rect(
        x_range: RangeInclusive<f64>,
        y_range: RangeInclusive<f64>,
        rect: Rect,
    ) -> Self {
        let scale_x = (rect.max.x - rect.min.x) as f64 / (x_range.end() - x_range.start());
        let scale_y = (rect.max.y - rect.min.y) as f64 / (y_range.end() - y_range.start());
        let offs_x = rect.min.x as f64 - scale_x * x_range.start();
        let offs_y = rect.min.y as f64 - scale_y * y_range.start();

        GraphTransform {
            scale_x,
            scale_y,
            offs_x,
            offs_y,
        }
    }
}

impl Mul<(f64, f64)> for GraphTransform {
    type Output = Pos2;

    fn mul(self, (in_x, in_y): (f64, f64)) -> Self::Output {
        let x = self.scale_x * in_x + self.offs_x;
        let y = self.scale_y * in_y + self.offs_y;
        Pos2::new(x as f32, y as f32)
    }
}

impl Graph {
    pub fn draw(
        ui: &mut Ui,
        graph_request: &GraphRequest,
        render_hint: &GraphRenderingHints,
        reader: &MetricsLogReadingRequester,
    ) {
        let context_menu_id = ui.id().with("context menu");

        EguiFrame::dark_canvas(ui.style()).show(ui, |ui| {
            // if wanted, we could do this:
            // ui.ctx().request_repaint();

            // store scaling stuff in here?
            //ui.memory().data.

            let desired_size = ui.available_width() * Vec2::new(1.0, 0.35);
            let response = ui.allocate_response(desired_size, Sense::click_and_drag());

            if response.clicked_by(PointerButton::Secondary) {
                ui.memory().open_popup(context_menu_id);
            }

            // not clickable:
            // egui::popup::show_tooltip_at(ui.ctx(), context_menu_id, Some(Pos2::new(320.0, 320.0)), |ui| {
            //     ui.button("hrm?");
            // });

            egui::popup_below_widget(ui, context_menu_id, &response, |ui| {
                ui.button("Blah");
            });

            let display_rect = response.rect;

            let current_window = reader.shared.current_window.read().unwrap();

            let mut x_axis = *current_window.time_range.start()..=*current_window.time_range.end();
            if x_axis.end() == x_axis.start() {
                x_axis = *x_axis.start()..=*x_axis.start() + 1.0;
            }

            // This range is reversed because screen coordinates go down, but we'd like them to go
            // up since this is more of a mathematical graph.
            let mut y_axis = if let Some(histogram) = current_window.histograms.get(graph_request) {
                histogram.y_axis.clone().into_inner()
            } else if let Some(scalar) = current_window.scalars.get(graph_request) {
                scalar.y_axis.clone().into_inner()
            } else {
                (0.0, 10.0)
            };

            if let Some((a, b)) = render_hint.include_y_range {
                // Force the y-axis to contain a range. Useful if you like 0 to be anchored on a
                // rate graph or something like that.
                if y_axis.0 > a {
                    y_axis.0 = a;
                }
                if y_axis.1 < b {
                    y_axis.1 = b;
                }
            }

            let y_axis = (y_axis.1)..=(y_axis.0);

            // let display_transform =
            //     emath::RectTransform::from_to(Rect::from_x_y_ranges(x_axis, y_axis), display_rect);

            let display_transform = GraphTransform::from_ranges_and_display_rect(
                x_axis.clone(),
                y_axis.clone(),
                display_rect,
            );

            let stroke = Stroke::new(2.0, Color32::GREEN);

            let mut last_point = None;

            let ticks = axis_ticks::ticks(y_axis.start().clone(), y_axis.end().clone(), 10);
            let tick_stroke = Stroke {
                width: 0.5,
                color: Color32::LIGHT_GRAY,
            };
            for tick in ticks {
                let text = ui.fonts().layout_no_wrap(
                    format!("{}", tick),
                    TextStyle::Small,
                    Color32::LIGHT_GRAY,
                );

                let mut text_rect = text.rect;
                text_rect.set_center(display_transform * (*x_axis.start(), tick));
                text_rect.set_left(display_rect.left());
                ui.painter().line_segment(
                    [
                        display_transform * (*x_axis.start(), tick)
                            + Vec2::new(5.0 + text.rect.width(), 0.0),
                        display_transform * (*x_axis.end(), tick),
                    ],
                    tick_stroke,
                );

                ui.painter().galley(text_rect.left_top(), text);
            }

            if let Some(scalar) = current_window.scalars.get(graph_request) {
                for (x, y) in scalar.points.iter() {
                    let new_point = display_transform * (x.as_f64_seconds(), *y);
                    ui.painter().circle_filled(new_point, 5.0, Color32::RED);
                    if let Some(last_point) = last_point {
                        ui.painter().line_segment([last_point, new_point], stroke);
                    }
                    last_point = Some(new_point);
                }
            }

            if let Some(histogram) = current_window.histograms.get(graph_request) {
                let min_density = histogram.min_density;
                let max_density = histogram.max_density;
                for heatrect in histogram.heatmap.iter() {
                    let rect = Rect::from_min_max(
                        display_transform * (heatrect.min_ts.as_f64_seconds(), heatrect.max_val),
                        display_transform * (heatrect.max_ts.as_f64_seconds(), heatrect.min_val),
                    );
                    eprintln!("R {:?} D {:?}", rect, display_rect);

                    let colour_cold = Color32::LIGHT_YELLOW;
                    let colour_hot = Color32::RED;

                    let lerp =
                        ((heatrect.density - min_density) / (max_density - min_density)) as f32;
                    eprintln!(
                        "Lerp {:?} D {:?} mD {:?} MD {:?}",
                        lerp, heatrect.density, min_density, max_density
                    );

                    // TODO refactor this to be a ColorF32 thing or something?
                    let (cr, cg, cb, ca) = colour_cold.to_tuple();
                    let (hr, hg, hb, ha) = colour_hot.to_tuple();
                    let mr = ((hr as f32 - cr as f32) * lerp + cr as f32) as u8;
                    let mg = ((hg as f32 - cg as f32) * lerp + cg as f32) as u8;
                    let mb = ((hb as f32 - cb as f32) * lerp + cb as f32) as u8;
                    let ma = ((ha as f32 - ca as f32) * lerp + ca as f32) as u8;

                    let lerp_colour = Color32::from_rgba_premultiplied(mr, mg, mb, ma);

                    eprintln!("LC {:?}", lerp_colour);

                    ui.painter().rect_filled(rect, 5.0, lerp_colour);
                }
                /*
                    // for x in hist.iter_all() {
                    //     debug!("H {:?}", x);
                    // }

                    let min_val = hist.min();
                    let max_val = hist.max();

                    let mut prev = min_val;
                    for q in (1..=10).map(|i| i as f64 * 0.1) {
                        // TODO improve this. It's going to be a bit off because of the way that
                        // numbers can be 'equivalent' etc ...
                        let value_at_q = hist.value_at_quantile(q);
                        let count_between = hist.count_between(prev, value_at_q);
                        let area = ((value_at_q - prev) * (xb.0 - xa.0)) as f64;
                        let density = count_between as f64 / area;
                        //eprintln!("den {}", density);
                        // TODO need to autoscale based on density...

                        //debug!("xa{} xb{}", xa.as_f64_seconds(), xb.as_f64_seconds());
                        let rect = Rect::from_min_max(
                            display_transform * (xa.as_f64_seconds(), value_at_q as f64),
                            display_transform * (xb.as_f64_seconds(), prev as f64)
                        );

                        ui.painter().rect_filled(rect, 5.0, Color32::DEBUG_COLOR);
                    }

                    //debug!("hist min {:?} max {:?}", hist.min(), hist.max());

                    // let minimum = hist.value_at_quantile(0.0);
                    // hist.max
                }
                */
            }
        });
    }
}
