use crate::background_loader::{
    MetricsLogReaderMessage, MetricsLogReaderNotification, MetricsLogReadingRequester,
};
use crate::config::{DashboardConfig, GraphRenderingHints, GraphRequest};
use crate::graph::Graph;

use bare_metrics_reader::MetricsLogReader;
use eframe::egui::{CentralPanel, CtxRef};
use eframe::epi::{App, Frame, Storage};
use eframe::NativeOptions;
use env_logger::Env;
use log::{debug, info, warn};
use std::collections::BTreeMap;
use std::fs::File;
use std::path::PathBuf;
use std::str::FromStr;

pub mod background_loader;
pub mod config;
pub mod graph;

pub struct MetricsGui {
    requester: MetricsLogReadingRequester,

    dashboard: Option<DashboardConfig>,
    // TODO we may want multiple graphs on the same pane, somehow.
    graph_requests: Vec<(GraphRequest, GraphRenderingHints)>,
}

impl App for MetricsGui {
    fn update(&mut self, ctx: &CtxRef, _frame: &mut Frame<'_>) {
        self.handle_messages();

        let Self {
            requester,
            dashboard,
            graph_requests,
        } = self;
        CentralPanel::default().show(ctx, |ui| {
            egui::ScrollArea::new([false, true]).show(ui, |ui| {
                if let Some(_dashboard) = dashboard {
                    let window = requester.shared.current_window.read().unwrap();
                    for (request, render_hint) in graph_requests.iter() {
                        ui.label(&request.metric_name);

                        if window.scalars.contains_key(request)
                            || window.histograms.contains_key(request)
                        {
                            Graph::draw(ui, request, render_hint, requester);
                        } else {
                            // TODO clarify
                            ui.label("(loading or missing ...)");
                        }
                    }
                } else {
                    ui.label("(no dashboard)");
                    // TODO Graph::draw(ui, MetricId(0), &requester);
                }
            });
        });
    }

    fn setup(&mut self, ctx: &CtxRef, _frame: &mut Frame<'_>, _storage: Option<&dyn Storage>) {
        let ctx = ctx.clone();

        *(self.requester.hook.lock().unwrap()) = Some(Box::new(move || {
            ctx.request_repaint();
        }));

        self.requester
            .tx
            .send(MetricsLogReaderMessage::LoadNewWindow {
                new_time_range: f64::NEG_INFINITY..=f64::INFINITY,
                new_wanted_time_points: 512,
                requests: Vec::with_capacity(0),
            })
            .unwrap();
    }

    fn name(&self) -> &str {
        "Bare Metrics GUI"
    }
}

impl MetricsGui {
    fn handle_messages(&mut self) {
        while let Ok(msg) = self.requester.rx.try_recv() {
            match msg {
                MetricsLogReaderNotification::MetricDescriptorsAvailable { .. } => {
                    info!("Metric descriptors are available");
                    self.on_metric_descriptors_available();
                }
                MetricsLogReaderNotification::WindowLoadStateChange { .. } => {
                    debug!("State change");
                }
            }
        }
    }

    fn on_metric_descriptors_available(&mut self) {
        if !self.graph_requests.is_empty() {
            warn!("TODO: don't know how to handle multiple appearances of descriptors");
            return;
        }
        if let Some(dashboard) = self.dashboard.as_ref() {
            for graph in dashboard.graphs.iter() {
                // TODO frankly we should maybe be using metric IDs in the graph requests
                self.graph_requests.push((
                    GraphRequest {
                        metric_name: graph.name.clone(),
                        // TODO we need something better about the metric labels
                        metric_labels: BTreeMap::new(),
                        derivation: graph.transform.clone(),
                    },
                    graph.rendering.clone(),
                ));
            }
        } else {
            // TODO Build some stuff automatically based on alphabetical ordering
        }

        self.requester
            .tx
            .send(MetricsLogReaderMessage::LoadNewWindow {
                new_time_range: f64::NEG_INFINITY..=f64::INFINITY,
                new_wanted_time_points: 512,
                requests: self
                    .graph_requests
                    .iter()
                    .map(|(a, _)| a)
                    .cloned()
                    .collect(),
            })
            .unwrap();
    }
}

fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    let path = if let Some(path_arg) = &std::env::args().skip(1).next() {
        PathBuf::from_str(path_arg)?
    } else {
        rfd::FileDialog::new()
            .set_title("Open a Bare Metrics Log")
            .add_filter("Bare Metrics Log", &["baremetrics"])
            .pick_file()
            .expect("You must pick a file!")
    };

    let file = File::open(path).unwrap();

    let reader = MetricsLogReader::new(file).unwrap();
    let log_app_name = reader.header.application_name.clone();
    let requester = MetricsLogReadingRequester::new_manager(reader, None);

    let dashboard_dir = appdirs::user_data_dir(Some("baremetrics"), Some("rei"), true)
        .unwrap()
        .join("gui/dashboards");

    let filtered_log_app_name = log_app_name.to_lowercase().replace(' ', "_");
    let dashboard_file = dashboard_dir.join(filtered_log_app_name + ".ron");
    info!("Looking for dashboard file at {:?}", dashboard_file);

    let dashboard = if dashboard_file.exists() {
        let rules_data: DashboardConfig = ron::de::from_reader(File::open(dashboard_file)?)?;
        info!("Dashboard loaded successfully :)");
        Some(rules_data)
    } else {
        info!("No dashboard found.");
        None
    };

    let app = MetricsGui {
        requester,
        dashboard,
        graph_requests: vec![],
    };

    let native_options = NativeOptions::default();

    eframe::run_native(Box::new(app), native_options);
}
